/*

    File: Main.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <assert.h>
#include <windows.h>
#include <gl/gl.h>

#include "../m_system3d/main.h"
#include "../m_foundation/main.h"
#include "../m_foundation/types.h"

#include "main.h"
#include "menu.h"

static int menu_count;

static CVarw cv_menu_count(&menu_count,		"menu_count");

////types.h
typedef struct {
	float xmin, xmax, ymin, ymax;
} rectf_t;

//*********************************
//menus
//*********************************

class menu_root_t : public menu_t {
public:
	menu_root_t::menu_root_t() : menu_t(0,0,MENU_X_MAX,MENU_Y_MAX) {
		color = 0;
	}
};

static menu_root_t *rootmenu = NULL;
static menu_t *deadmenu = NULL;		//where we link the Destroy()'d ones to-be-deleted

void menuUnlinkFromParent(menu_t *menu) {
	if (!menu->parent) {	//if no parent then no neighbors
		assert(!menu->prev && !menu->next);
		return;
	}
	if (menu->parent->child == menu) {		//first child - no prev neighbor
		assert(!menu->prev);
		menu->parent->child = menu->next;
		if (menu->next) menu->next->prev = NULL;
	} else {
		assert(menu->prev);
		menu->prev->next = menu->next;
		if (menu->next) menu->next->prev = menu->prev;
	}
	menu->parent = menu->prev = menu->next = NULL;
}


void menuAddChild(menu_t *parent, menu_t *child) {
	assert(child);
	menuUnlinkFromParent(child);
	child->parent = parent;
	if (!parent->child) {
		parent->child = child;
	} else {
		for (menu_t *m = parent->child; m->next; m = m->next);
		m->next = child;
		child->prev = m;
	}
}

void menuInit(menu_t *m, float x, float y, float sx, float sy) {
	//ctor
	m->active = true;
	m->parent = NULL;
	m->child = NULL;
	m->prev = NULL;
	m->next = NULL;
	m->pos[0] = x;
	m->pos[1] = y;
	m->color = 0xFFFFFFFF;
	m->texture = 0;
	m->childofs[0] = 0;
	m->childofs[1] = 0;
	m->topmost_priority = 0;
	memset(m->text, 0, sizeof(m->text));
	memset(m->identifier, 0, sizeof(m->identifier));
	//Init code
	m->pos[0] = x;
	m->pos[1] = y;
	m->setSize(sx, sy);
}


void menuDelete(menu_t *menu) {
	menu->active = false;
	menuUnlinkFromParent(menu);
	while (menu->child) menuDelete(menu->child);
	free(menu);
}

void menuDestroy(menu_t *menu) {
	menuUnlinkFromParent(menu);
	if (!deadmenu) {
		deadmenu = menu;
	} else {
		menuAddChild(deadmenu, menu);
	}
}

menu_t *menuFind(menu_t *parent, const char *ident) {
	//found a match
	if (!strncmp(ident, parent->identifier, sizeof(parent->identifier))) {
		return parent;
	}

	//search children
	for (menu_t *c = parent->child; c; c = c->next) {
		menu_t *result = menuFind(c, ident);
		if (result) return result;
	}

	//no luck
	return NULL;
}

void Menu_SetTopmost(menu_t *menu) {

	//place it at the beginning of the root children
	if (!menu->next) return;		//if 'next' is set then we are not yet last child

	//unlink

	//unlink from top
	if (menu->parent->child == menu) {
		menu->parent->child = menu->next;
	}
	//unlink from left
	if (menu->prev) {
		menu->prev->next = menu->next;
	}
	//unlink from right
	menu->next->prev = menu->prev;
	//cycle to end
	for (menu_t *last = menu->next; last->next; last = last->next);
	//relink to left
	last->next = menu;
	menu->prev = last;
	//clear right link
	menu->next = NULL;
}

bool Menu_IsUnderCoord(menu_t *menu, float x, float y) {
	x -= menu->pos[0];
	y -= menu->pos[1];
	return (x > 0 && y > 0 && x < menu->size[0] && y < menu->size[1]);
}

bool Menu_ProcessMouseClick(menu_t *menu, int event, float x, float y) {
	//assume incoming coordinates are relative to 'parent'
	//convert to be relative to 'this'
	x -= menu->pos[0];
	y -= menu->pos[1];

	float child_x = x + menu->childofs[0];
	float child_y = y + menu->childofs[1];

	menu->selectedChild = NULL;

	//handle input for all children
	//chain last = topmost = first to handle input
	menu_t *c = menu->child;
	if (c) {
		for (; c->next; c = c->next);	//cycle to the end of the list
		menu_t *prev;
		for (; c; c = prev) {		//then go backwards
			prev = c->prev;	//just in case we get some unlinking...

			//if c is under <x,y> then execute MouseLPress(x,y), select the object
			if (Menu_IsUnderCoord(c, child_x, child_y)) {
				menu->selectedChild = c;
				if (Menu_ProcessMouseClick(c, event, child_x, child_y)) return true;
			}
		}
	}

	return menu->doMouseClick(event, x,y);
}

//*********************************
//menu display
//*********************************

void Menu_Display(menu_t *menu, rectf_t rect) {

	menu_count++;

	//convert 'rect' to relative coordinate space
	rect.xmin -= menu->pos[0];
	rect.xmax -= menu->pos[0];
	rect.ymin -= menu->pos[1];
	rect.ymax -= menu->pos[1];

	//check relative clipping area
		//occlude
	if (rect.xmin > menu->size[0]) return;
	if (rect.xmax < 0) return;
	if (rect.ymin > menu->size[1]) return;
	if (rect.ymax < 0) return;
		//clamp
	if (rect.ymin < 0) rect.ymin = 0;
	if (rect.ymax > menu->size[1]) rect.ymax = menu->size[1];
	if (rect.xmin < 0) rect.xmin = 0;
	if (rect.xmax > menu->size[0]) rect.xmax = menu->size[0];


	//render current menu

	glPushMatrix();
	glTranslatef(menu->pos[0], menu->pos[1], 0);

//assume glEnable(GL_TEXTURE_2D);

	//set up our clipping planes in the relative space
	double plane[4] = {0,0,0,0};
	//top
	plane[0] = 0;
	plane[1] = 1;
	plane[2] = 0;
	plane[3] = -rect.ymin;
	glClipPlane(GL_CLIP_PLANE0, plane);
	//bottom
	plane[0] = 0;
	plane[1] = -1;
	plane[2] = 0;
	plane[3] = rect.ymax;
	glClipPlane(GL_CLIP_PLANE1, plane);
	//left
	plane[0] = 1;
	plane[1] = 0;
	plane[2] = 0;
	plane[3] = -rect.xmin;
	glClipPlane(GL_CLIP_PLANE2, plane);
	//right
	plane[0] = -1;
	plane[1] = 0;
	plane[2] = 0;
	plane[3] = rect.xmax;
	glClipPlane(GL_CLIP_PLANE3, plane);

	glEnable(GL_CLIP_PLANE0);
	glEnable(GL_CLIP_PLANE1);
	glEnable(GL_CLIP_PLANE2);
	glEnable(GL_CLIP_PLANE3);

	//render

	glColor4ubv((unsigned char *)(&menu->color));

	if (menu->color & 0xFF000000) {	//no render if fully transparent

		if (menu->texture) {
			glBindTexture(GL_TEXTURE_2D, menu->texture);
		} else {
			glDisable(GL_TEXTURE_2D);
		}
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(0,				0);
		glTexCoord2f(0,1);
		glVertex2f(0,				menu->size[1]);
		glTexCoord2f(1,1);
		glVertex2f(menu->size[0],	menu->size[1]);
		glTexCoord2f(1,0);
		glVertex2f(menu->size[0],	0);
		glEnd();
		if (!menu->texture) {
			glEnable(GL_TEXTURE_2D);
		}
	}

	if (menu->text) {
		main_console->font->Set();
		int len = strlen(menu->text);
		const char *p = menu->text;
		for (int a = 0; a < len; a++, p++) {
			int tx = (*(unsigned char *)p & 15);
			int ty = (*(unsigned char *)p >> 4) - 2;
			float vtx[4][2] = {	{0,	0},	{0,	1},	{1,	1},	{1,	0}};
			glBegin(GL_QUADS);
			for (int i = 0; i < 4;i ++) {
				glTexCoord2f((float)(tx+vtx[i][0])/16.f, (float)(ty+vtx[i][1])/16.f);
				glVertex2f(vtx[i][0] + a,vtx[i][1]);
			}
			glEnd();
		}
	}

	glDisable(GL_CLIP_PLANE0);
	glDisable(GL_CLIP_PLANE1);
	glDisable(GL_CLIP_PLANE2);
	glDisable(GL_CLIP_PLANE3);

	//used especially for child clipping regions =) virtual-to-real world mapping =)
	glTranslatef(-menu->childofs[0], -menu->childofs[1], 0);

	rect.xmin += menu->childofs[0];
	rect.xmax += menu->childofs[0];
	rect.ymin += menu->childofs[1];
	rect.ymax += menu->childofs[1];

	//render all children
	for (menu_t *c = menu->child; c; c = c->next) {
		//prior to each child display, if we are clipping children then re-init clip planes 
		//this must be done once each render because child nodes could potentially
		//set up the clip planes themselves

		Menu_Display(c, rect);
	}

	glPopMatrix();
}

//override the ProcessMouseClick - add in our 'put the last clicked on top' ability
bool MenuRoot_ProcessMouseClick(menu_t *m, int event, float x, float y) {
	bool result = Menu_ProcessMouseClick(m, event,x,y);

	if (m->selectedChild) {
		Menu_SetTopmost(m->selectedChild);
	}

	return result;
}

//do the geometric convergence
void menuSystemDisplay(void) {

	glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,MENU_X_MAX,MENU_Y_MAX,0,-1000,1000);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	//prep our viewport - disable

	//render recursively all non-parented nodes 
	if (rootmenu &&
		rootmenu->active) {
		glEnable(GL_BLEND);

		menu_count = 0;	//zero the counter before accumulating - for debug

		rectf_t rect;
		rect.xmin = 0;
		rect.xmax = MENU_X_MAX;
		rect.ymin = 0;
		rect.ymax = MENU_Y_MAX;
		Menu_Display(rootmenu, rect);

		glDisable(GL_BLEND);
	}

	glPopMatrix();		//pop the modelview matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();		//pop the projection matrix
	glMatrixMode(GL_MODELVIEW);	//and set the current matrix back to modelview

	glEnable(GL_DEPTH_TEST);
}

//*********************************
//menu input
//*********************************

bool menuSystemGetInput(void) {

	int event_flags = 0;
	if (main_mouse->buttonState[0] == MH_PRESS)	event_flags |= (1 << MENUFUNC_LPRESS);
	if (main_mouse->buttonState[1] == MH_PRESS)	event_flags |= (1 << MENUFUNC_RPRESS);	
	if (main_mouse->buttonState[0] == MH_DOWN)	event_flags |= (1 << MENUFUNC_LDOWN);
	if (main_mouse->buttonState[1] == MH_DOWN)	event_flags |= (1 << MENUFUNC_RDOWN);	
	if (!event_flags) return false;	//no click yet...

	float mouse_x = main_mouse->mouseSX * MENU_X_MAX;
	float mouse_y = (1.f - main_mouse->mouseSY) * MENU_Y_MAX;

	if (!rootmenu || !rootmenu->active) return false;

	//bool result = rootmenu->MouseClick(event_flags, mouse_x, mouse_y);
	bool result = MenuRoot_ProcessMouseClick(
		(menu_t *)rootmenu, event_flags, mouse_x, mouse_y);

#if 0
	//resort by topmost_priority:
	//somewhere in here we're losing links in the chain
	if (rootmenu->child) {
		CGameMenu *m = rootmenu->child->next;
		if (m) {
			m->prev = NULL;
			rootmenu->child->next = NULL;
			//now we're dealing with 'rootmenu' with at least two children
			CGameMenu *next = NULL;
			for (; m; m = next) {
				//grab our next guy
				next = m->next;
				if (next) next->prev = NULL;	//if he exists then unlink his prev
				//unlink the current from prev/next.  dont touch parent - this wont change
				assert(!m->prev);
				m->next = NULL;
				//then cycle thru the new parent, relinking the current
				CGameMenu *child = rootmenu->child;
				CGameMenu *lastchild = 0;
				for (; child; child = child->next) {
					lastchild = child;
					if (child->topmost_priority < m->topmost_priority) break;
				}
				if (child) {
					m->prev = child->prev;
					if (m->prev) m->prev->next = m;
					else rootmenu->child = m;
					child->prev = m;
					m->next = child;
					break;
				} else {
					assert(lastchild);
					m->prev = lastchild;
					lastchild->next = m;
				}
			}
		}
	}
#endif

	//do this once per frame - anywhere
	if (deadmenu) {
		menuDelete(deadmenu);
		deadmenu = NULL;
	}

	return result;
}

menu_t *menuGetRoot(void) {
	return rootmenu;
}

//*********************************
//menu system
//*********************************

bool menuSystemInit(void) {
	//wipe them all

	rootmenu = new menu_root_t();

	return true;
}

void menuSystemShutdown(void) {
	//dont use 'delete', because that is script-wired to just link the menu to the flush buffer
	if (rootmenu) free(rootmenu);
	rootmenu = NULL;
}
